/*
 * This program is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published 
 * by the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU Library General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
 */

#include "Sort.h"
#include <errno.h>
#include <sys/types.h>
#include <sys/time.h>
#include <unistd.h>
#include <assert.h>

@implementation Sort

+ new
{
    self = [super new];
    compareVal = 1;
    moveVal = 2;
    fcallVal = 8;
    animate = NO;
    return self;
}

/* send by parent after 'receive' returns nil */

- free
{
  if (close(drawfd) != 0) perror("close"); /* widow pipe at read end */
  if (close(schedfd) != 0) perror("close"); /* widow sched pipe at write end */
  return [super free];
}

- (int)fork
{
  int fd[2];
  int sfd[2];

  if (pipe(fd) == -1) { perror("pipe");return -1; }
  if (pipe(sfd) == -1) { perror("pipe");return -1; }

  pid = fork();

  if (pid == -1) { perror("fork");return -1; }

  if (pid == 0) {
    close(fd[0]);
    close(sfd[1]);
    drawfd = fd[1];
    schedfd = sfd[0];
    [self sort];
    _exit(0); /* widow drawfd pipe at write end */
  } else {
    drawfd = fd[0];
    schedfd = sfd[1];
    close(fd[1]);
    close(sfd[0]);
  }

  return pid;
}

- (int)drawfd
{
  return drawfd;
}

- send:(sortmsg *)m
{
  if (write(drawfd,m,sizeof(sortmsg)) != sizeof(sortmsg)) {
    perror("write");
  }
  return self;
}

- receive:(sortmsg *)m
{
  if (read(drawfd,m,sizeof(sortmsg)) != sizeof(sortmsg)) {
    return nil; /* EOF */
  } else {
    m->who = self;
    return self;
  }
}

- wakeup:(BOOL)abort
{
  if (pid) {
    if (write(schedfd,&abort,sizeof(char)) != sizeof(char)) {
      perror("write");
    }
  }
  return self;
}

- wakeup
{
  [self wakeup:NO];
  return self;
}

- abort
{
  if (pid) {
    [self wakeup:YES];
    return self;
  } else {
    [self sortDone];
    _exit(0); /* widow drawfd write end of pipe */
    return nil;
  }
}


- block
{
  int n;
  sortmsg msg;
  char abortflag;

  msg.what = TICK_MSG;
  msg.value[0] = [self totalTicks];
  [self send:&msg];

  n = read(schedfd,&abortflag,sizeof(char));
  if (n == 0) _exit(0);
  if (n != sizeof(char)) perror("read");

  if (abortflag) {
    [self abort];
  }

  return self;
}

- setticks:(int)value for:(int)tag
{
#define COMPARE 3
#define MOVE 4
#define FCALL 5	

    switch (tag) {
        case COMPARE:	compareVal = value; 
			break;
        case MOVE: 	moveVal = value;
		 	break;
        case FCALL: 	fcallVal = value;
			break;
    }
    return self;
}

- setanimate:(BOOL)value
{
    animate = value;
    return self;
}

- setsize:(int)size address:(int*)address
{
    int i;
    numElements = size;
    data = (int *)calloc(numElements,sizeof(int));
    for(i=0;i<size;i++) {
      data[i] = address[i];
    }
    return self;
}


- sortDone
{
    sortmsg m; 
    m.what = FINISHED_MSG;    /* send finished message to parent */
    [self send:&m]; /* draw finished */
    if (data) free(data);		/* free the data set */
    numElements = 0;
    return self;
}

- adjust      
{
    [self block];
    return self;
}

- swap:(int)position1 with:(int)position2
{    
    int temp;
    sortmsg m;
    
    moves += 2;     /* a swap involves two moves */
    [self adjust];  
    m.what = SWAP_MSG;	/* send a swap message to parent */
    m.position[0] = position1;
    m.value[0] = data[position1];
    m.position[1] = position2;
    m.value[1] = data[position2];
    [self send:&m];
    temp = data[position1];
    data[position1] = data[position2];
    data[position2] = temp;
    return self;
}

- moveValue:(int)value to:(int)position
{      
    sortmsg m;
    moves++;
    [self adjust]; 
    m.what = MOVE_MSG;	/* send a move message to parent */
    m.value[0] = value;
    m.position[0] = position;
    m.value[1] = data[position];
    [self send:&m];
    data[position] = value;  
    return self;
}

- (BOOL)lessThan:(int)position1 :(int)position2
{        
    compares++;     
    [self adjust];
    if (animate) {	/* if animating compares */
        sortmsg m;
	m.what = COMPARE_MSG; /* send compare message to parent */
	m.position[0] = position1;
	m.value[0] = data[position1];
	m.position[1] = position2;
	m.value[1] = data[position2];
	[self send:&m];    
    }
    return (data[position1] < data[position2]);
}

- sort
{
    int c;
    sortmsg m;

    compares = moves = fcalls = 0;
    m.what = SORTING_MSG;
    [self send:&m];  /* status "Sorting" */

    for (c = 0; c < numElements; c++) {	/* for each element in the data set */
        sortmsg m;
        m.what = MOVE_MSG;
        m.position[0] = c;
        m.value[0] = data[c];
        m.value[1] = 0;
        [self send:&m]; 
    }
    return self;
}

- (const char*)getName
{
    return sortName;
}
- (int)compares
{
    return compares;
}
- (int)moves
{
    return moves;
}
- (int)fcalls
{
    return fcalls;
}
- (int)totalTicks
{
    return (compareVal*compares + moveVal*moves + fcallVal*fcalls);
}

@end

